#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function
from unittest import TestCase
from lingvoreader import bitstream

__author__ = 'sv99'


class TestBitStream(TestCase):
    def setUp(self):
        print("Creating a new BitStream...")
        # record - bytearray, not string!
        self.record = bytearray('\x00\x01\x02\x03\x04\x05\x06\x07\x08')
        self.bst = bitstream.BitStream(self.record)

    def tearDown(self):
        print("Destroying the BitStream...")
        self.bst = None

    def test_reverse32(self):
        self.assertEqual(bitstream.reverse32(0x01020304), 0x04030201)

    def test_reverse16(self):
        self.assertEqual(bitstream.reverse16(0x0102), 0x0201)

    def test_length(self):
        self.assertEqual(len(self.record), self.bst.length)

    def test_to_nearest_byte(self):
        self.bst.seek(0)
        self.bst.to_nearest_byte()
        self.assertEqual(self.bst.pos, 0)
        self.bst.read_bit()
        self.bst.to_nearest_byte()
        self.assertEqual(self.bst.pos, 1)

    def test_read_byte(self):
        self.bst.seek(0)
        self.assertEqual(self.bst.read_byte(), 0)
        self.bst.seek(1)
        self.assertEqual(self.bst.read_byte(), 1)

    def test_read_word(self):
        self.bst.seek(1)
        self.assertEqual(self.bst.read_word(), 0x0102)

    def test_read_int(self):
        self.bst.seek(1)
        self.assertEqual(self.bst.read_int(), 0x01020304)

    #def test_read_symbols(self):
    #    self.fail()

    def test_read_bit(self):
        self.bst.seek(1)
        self.bst.read_bits(7)
        self.assertEqual(self.bst.read_bit(), 1)
        self.assertEqual(self.bst.read_bit(), 0)

    def test_read_bits(self):
        self.bst.seek(1)
        self.assertEqual(self.bst.read_bits(4), 0)
        self.assertEqual(self.bst.read_bits(4), 1)
        self.assertEqual(self.bst.read_bits(4), 0)
        self.assertEqual(self.bst.read_bits(8), 0x20)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (division, absolute_import, print_function)
from lingvoreader import tools
from lingvoreader.lentable import LenTable
from lingvoreader.tools import int2unichr

__author__ = 'sv99'


class Decoder:
    def __init__(self, bstr):
        self.bstr = bstr
        self.prefix = ""
        self._article_symbols = None
        self._heading_symbols = None
        self._ltArticles = None
        self._ltHeadings = None
        self._ltPrefixLengths = None
        self._ltPostfixLengths = None
        self._huffman1Number = 0
        self._huffman2Number = 0
        self._readed = False

    @property
    def readed(self):
        return self._readed

    def decode_prefix_len(self):
        return self._ltPrefixLengths.decode()

    def decode_postfix_len(self):
        return self._ltPostfixLengths.decode()

    def read_reference1(self):
        return self.read_reference(self._huffman1Number)

    def read_reference2(self):
        return self.read_reference(self._huffman2Number)

    def read_reference(self, huffman_number):
        reference = ""
        code = self.bstr.read_bits(2)
        if code == 3:
            self.bstr.read_bits(32)
            return reference

        size = tools.bit_length(huffman_number)
        assert(size >= 2)
        return (code << (size - 2)) | self.bstr.read_bits(size - 2)

    def decode_heading(self, size):
        res = ""
        for i in range(size):
            sym_idx = self._ltHeadings.decode()
            sym = self._heading_symbols[sym_idx]
            assert(sym <= 0xffff)  # LingvoEngine:2EAB84E8
            res += int2unichr(sym)
        return res

    def decode_article(self, size):
        """
        decode User and Abrv dict
        """
        res = ""
        while len(res) < size:
            sym_idx = self._ltArticles.decode()
            sym = self._article_symbols[sym_idx]
            if sym >= 0x10000:
                if sym >= 0x10040:
                    start_idx = self.bstr.read_bits(tools.bit_length(size))
                    s = sym - 0x1003d
                    res += res[start_idx:start_idx + s]
                else:
                    prefix_idx = self.bstr.read_bits(tools.bit_length(len(self.prefix)))
                    s = sym - 0xfffd
                    res += self.prefix[prefix_idx:prefix_idx + s]
            else:
                res += int2unichr(sym)
        return res

    # need seek(bstr.header.dictionary_encoder_offset) befor call!
    def read(self):
        return

    def dump(self):
        print("Decoder:               %s" % self.__class__.__name__)
        if self.readed:
            print("    ArticleSymbols:    %d" % len(self._article_symbols))
            print("    HeadingSymbols:    %d" % len(self._heading_symbols))
            self._ltArticles.dump("Articles")
            self._ltHeadings.dump("Headings")
            self._ltPrefixLengths.dump("PrefixLengths")
            self._ltPostfixLengths.dump("PostfixLengths")


class UserDictionaryDecoder(Decoder):
    def __init__(self, bstr):
        Decoder.__init__(self, bstr)
        return

    def read(self):
        self.prefix = self.bstr.read_unicode(self.bstr.read_int())
        self._article_symbols = self.bstr.read_symbols()
        self._heading_symbols = self.bstr.read_symbols()
        self._ltArticles = LenTable(self.bstr)
        self._ltHeadings = LenTable(self.bstr)

        self._ltPrefixLengths = LenTable(self.bstr)
        self._ltPostfixLengths = LenTable(self.bstr)

        self._huffman1Number = self.bstr.read_bits(32)
        self._huffman2Number = self.bstr.read_bits(32)
        self._readed = True
        return


class SystemDictionaryDecoder13(Decoder):
    def __init__(self, bstr):
        Decoder.__init__(self, bstr)
        return

    def read(self):
        self.prefix = self.bstr.read_unicode(self.bstr.read_int())
        self._article_symbols = self.bstr.read_symbols()
        self._heading_symbols = self.bstr.read_symbols()
        self._ltArticles = LenTable(self.bstr)
        self._ltHeadings = LenTable(self.bstr)

        self._ltPrefixLengths = LenTable(self.bstr)
        self._ltPostfixLengths = LenTable(self.bstr)

        self._huffman1Number = self.bstr.read_bits(32)
        self._huffman2Number = self.bstr.read_bits(32)
        self._readed = True
        return

    def decode_article(self, size):
        res = ""
        while len(res) < size:
            sym_idx = self._ltArticles.decode()
            sym = self._article_symbols[sym_idx]
            if sym <= 0x80:
                if sym <= 0x3F:
                    start_pref_idx = self.bstr.read_bits(tools.bit_length(len(self.prefix)))
                    s = sym + 3
                    res += self.prefix[start_pref_idx:start_pref_idx + s]
                else:
                    start_idx = self.bstr.read_bits(tools.bit_length(size))
                    s = sym - 0x3d
                    res += res[start_idx:start_idx + s]
            else:
                res += int2unichr(sym - 0x80)
        return res


class SystemDictionaryDecoder14(SystemDictionaryDecoder13):
    def __init__(self, bstr):
        SystemDictionaryDecoder13.__init__(self, bstr)
        return

    def read(self):
        prefix_len = self.bstr.read_int()
        self.prefix = self.bstr.read_unicode(prefix_len)
        self._article_symbols = self.bstr.read_symbols()
        self._heading_symbols = self.bstr.read_symbols()
        self._ltArticles = LenTable(self.bstr)
        self._ltHeadings = LenTable(self.bstr)

        self._ltPostfixLengths = LenTable(self.bstr)
        self._dummy = self.bstr.read_bits(32)
        self._ltPrefixLengths = LenTable(self.bstr)

        self._huffman1Number = self.bstr.read_bits(32)
        self._huffman2Number = self.bstr.read_bits(32)
        self._readed = True
        return

    # def decode_article(self, size):
    #     res = ""
    #     while len(res) < size:
    #         sym_idx = self._ltArticles.decode()
    #         sym = self._article_symbols[sym_idx]
    #         if sym <= 0x80:
    #             if sym <= 0x3F:
    #                 start_pref_idx = self.bstr.read_bits(tools.bit_length(len(self.prefix)))
    #                 s = sym + 3
    #                 res += self.prefix[start_pref_idx:start_pref_idx + s]
    #             else:
    #                 start_idx = self.bstr.read_bits(tools.bit_length(size))
    #                 s = sym - 0x3d
    #                 res += res[start_idx:start_idx + s]
    #         else:
    #             res += unichr(sym - 0x80)
    #     return res


class AbbreviationDictionaryDecoder(Decoder):
    def __init__(self, bstr):
        Decoder.__init__(self, bstr)
        return

    def read(self):
        self.prefix = self.read_xored_prefix(self.bstr.read_int())
        self._article_symbols = self.read_xored_symbols()
        self._heading_symbols = self.read_xored_symbols()
        self._ltArticles = LenTable(self.bstr)
        self._ltHeadings = LenTable(self.bstr)

        self._ltPrefixLengths = LenTable(self.bstr)
        self._ltPostfixLengths = LenTable(self.bstr)

        self._huffman1Number = self.bstr.read_bits(32)
        self._huffman2Number = self.bstr.read_bits(32)
        self._readed = True
        return

    def read_xored_symbols(self):
        size = self.bstr.read_bits(32)
        bits_per_symbol = self.bstr.read_bits(8)
        res = []
        for i in range(size):
            res.append(self.bstr.read_bits(bits_per_symbol) ^ 0x1325)
        return res

    def read_xored_prefix(self, size):
        res = ""
        for i in range(size):
            res += int2unichr(self.bstr.read_bits(16) ^ 0x879A)
        return res


class SystemDictionaryDecoder15(Decoder):
    def __init__(self, bstr):
        Decoder.__init__(self, bstr)
        return

    def read(self):
        # self.bstr = XoredBitStream(self.bstr)
        # self.decode()

        prefix_len = self.bstr.read_some(4)
        self.prefix = self.bstr.read_unicode(prefix_len)
        self._article_symbols = self.bstr.read_symbols()
        self._heading_symbols = self.bstr.read_symbols()
        self._ltArticles = LenTable(self.bstr)
        self._ltHeadings = LenTable(self.bstr)

        self._ltPostfixLengths = LenTable(self.bstr)
        self._dummy = self.bstr.read_bits(32)
        self._ltPrefixLengths = LenTable(self.bstr)

        self._huffman1Number = self.bstr.read_bits(32)
        self._huffman2Number = self.bstr.read_bits(32)
        self._readed = True
        return

    def decode_article(self, size):
        res = ""
        while len(res) < size:
            sym_idx = self._ltArticles.decode()
            sym = self._article_symbols[sym_idx]
            if sym <= 0x80:
                if sym <= 0x3F:
                    start_pref_idx = self.bstr.read_bits(tools.bit_length(len(self.prefix)))
                    s = sym + 3
                    res += self.prefix[start_pref_idx:start_pref_idx + s]
                else:
                    start_idx = self.bstr.read_bits(tools.bit_length(size))
                    s = sym - 0x3d
                    res += res[start_idx:start_idx + s]
            else:
                res += int2unichr(sym - 0x80)
        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (print_function)
from builtins import chr as text

__author__ = 'sv99'


lang_map = {
    1555: u"Abazin",
    1556: u"Abkhaz",
    1557: u"Adyghe",
    1078: u"Afrikaans",
    1559: u"Agul",
    1052: u"Albanian",
    1545: u"Altaic",
    1025: u"Arabic",                # x5 tested support
    # 1025: u"ArabicSaudiArabia",
    5121: u"ArabicAlgeria",
    15361: u"ArabicBahrain",
    3073: u"ArabicEgypt",
    2049: u"ArabicIraq",
    11265: u"ArabicJordan",
    13313: u"ArabicKuwait",
    12289: u"ArabicLebanon",
    4097: u"ArabicLibya",
    6145: u"ArabicMorocco",
    8193: u"ArabicOman",
    16385: u"ArabicQatar",
    10241: u"ArabicSyria",
    7169: u"ArabicTunisia",
    14337: u"ArabicUAE",
    9217: u"ArabicYemen",
    1067: u"Armenian",              # x5 tested support
    # 1067: u"ArmenianEastern",
    33835: u"ArmenianGrabar",
    32811: u"ArmenianWestern",
    1101: u"Assamese",
    1558: u"Awar",
    1560: u"Aymara",
    2092: u"AzeriCyrillic",
    1068: u"AzeriLatin",
    1561: u"Bashkir",
    1069: u"Basque",
    1059: u"Belarusian",
    # 1059: u"Byelorussian",
    1562: u"Bemba",
    1093: u"Bengali",
    1563: u"Blackfoot",
    1536: u"Breton",
    1564: u"Bugotu",
    1026: u"Bulgarian",
    1109: u"Burmese",
    1565: u"Buryat",
    1027: u"Catalan",
    1566: u"Chamorro",
    1544: u"Chechen",
    1028: u"Chinese",               # x5 tested support
    # 1028: u"ChineseTaiwan",
    3076: u"ChineseHongKong",
    5124: u"ChineseMacau",
    2052: u"ChinesePRC",
    4100: u"ChineseSingapore",
    1567: u"Chukcha",
    1568: u"Chuvash",
    1569: u"Corsican",
    1546: u"CrimeanTatar",
    1050: u"Croatian",
    1570: u"Crow",
    1029: u"Czech",
    1030: u"Danish",
    1572: u"Dungan",
    1043: u"Dutch",                 # x5 tested
    2067: u"DutchBelgian",          # not supported
    1033: u"English",               # 1033: u"EnglishUnitedStates",
    3081: u"EnglishAustralian",
    10249: u"EnglishBelize",
    4105: u"EnglishCanadian",
    9225: u"EnglishCaribbean",
    6153: u"EnglishIreland",
    8201: u"EnglishJamaica",
    35849: u"EnglishLaw",
    33801: u"EnglishMedical",
    5129: u"EnglishNewZealand",
    13321: u"EnglishPhilippines",
    34825: u"EnglishProperNames",
    7177: u"EnglishSouthAfrica",
    11273: u"EnglishTrinidad",
    2057: u"EnglishUnitedKingdom",
    12297: u"EnglishZimbabwe",
    1573: u"EskimoCyrillic",
    1537: u"Esperanto",
    1061: u"Estonian",
    1574: u"Even",
    1575: u"Evenki",
    1065: u"Farsi",
    1538: u"Fijian",
    1035: u"Finnish",
    1036: u"French",                # x5 supported
    2060: u"FrenchBelgian",
    3084: u"FrenchCanadian",
    5132: u"FrenchLuxembourg",
    6156: u"FrenchMonaco",
    33804: u"FrenchProperNames",
    4108: u"FrenchSwiss",
    1122: u"Frisian",
    1576: u"Frisian_Legacy",  # x6
    1577: u"Friulian",
    1084: u"GaelicScottish",
    1578: u"Gagauz",
    1110: u"Galician",
    1579: u"Galician_Legacy",  # x6
    1580: u"Ganda",
    1079: u"Georgian",
    1031: u"German",
    3079: u"GermanAustrian",
    34823: u"GermanLaw",
    5127: u"GermanLiechtenstein",
    4103: u"GermanLuxembourg",
    36871: u"GermanMedical",
    32775: u"GermanNewSpelling",
    35847: u"GermanNewSpellingLaw",
    37895: u"GermanNewSpellingMedical",
    39943: u"GermanNewSpellingProperNames",
    38919: u"GermanProperNames",
    2055: u"GermanSwiss",
    1032: u"Greek",
    32776: u"GreekKathareusa",
    1140: u"Guarani",
    1582: u"Guarani_Legacy",  # x6
    1095: u"Gujarati",
    1583: u"Hani",
    1128: u"Hausa",  # x6
    1652: u"Hausa_Legacy",
    1141: u"Hawaiian",  # x6
    1539: u"Hawaiian_Legacy",
    1037: u"Hebrew",
    1081: u"Hindi",
    1038: u"Hungarian",
    1039: u"Icelandic",
    1584: u"Ido",
    1057: u"Indonesian",
    1585: u"Ingush",
    1586: u"Interlingua",
    2108: u"Irish",  # x6
    1552: u"Irish_Legacy",  # x6
    # 2108: u"Gaelic",          # x6
    # 1552: u"Gaelic_Legacy",   # x6
    1040: u"Italian",           # x5 tested
    33808: u"ItalianProperNames",
    2064: u"ItalianSwiss",
    1041: u"Japanese",
    1548: u"Kabardian",
    1587: u"Kalmyk",
    1099: u"Kannada",
    1589: u"KarachayBalkar",
    1588: u"Karakalpak",
    1120: u"Kashmiri",
    2144: u"KashmiriIndia",
    1590: u"Kasub",
    1591: u"Kawa",
    1087: u"Kazakh",
    1592: u"Khakas",
    1593: u"Khanty",
    1107: u"Khmer",
    1594: u"Kikuyu",
    1595: u"Kirgiz",
    1597: u"KomiPermian",
    1596: u"KomiZyryan",
    1598: u"Kongo",
    1111: u"Konkani",
    1042: u"Korean",
    2066: u"KoreanJohab",
    1599: u"Koryak",
    1600: u"Kpelle",
    1601: u"Kumyk",
    1602: u"Kurdish",
    1603: u"KurdishCyrillic",
    1604: u"Lak",
    1108: u"Lao",
    1142: u"Latin",  # x6
    1540: u"Latin_Legacy",
    1062: u"Latvian",
    1655: u"LatvianGothic",
    1605: u"Lezgin",
    1063: u"Lithuanian",
    2087: u"LithuanianClassic",
    1606: u"Luba",
    1071: u"Macedonian",
    1607: u"Malagasy",
    1086: u"Malay",
    2110: u"MalayBruneiDarussalam",
    1100: u"Malayalam",
    1608: u"Malinke",
    1082: u"Maltese",
    1112: u"Manipuri",
    1609: u"Mansi",
    1153: u"Maori",  # x6
    1102: u"Marathi",
    1610: u"Mari",
    1611: u"Maya",
    1612: u"Miao",
    1613: u"Minankabaw",
    1614: u"Mohawk",
    1104: u"Mongol",
    1615: u"Mordvin",
    1616: u"Nahuatl",
    1617: u"Nanai",
    1618: u"Nenets",
    1121: u"Nepali",
    2145: u"NepaliIndia",
    1619: u"Nivkh",
    1620: u"Nogay",
    1044: u"Norwegian",
    # 1044: u"NorwegianBokmal",
    2068: u"NorwegianNynorsk",
    1621: u"Nyanja",
    1622: u"Occidental",
    1623: u"Ojibway",
    32777: u"OldEnglish",
    32780: u"OldFrench",
    33799: u"OldGerman",
    32784: u"OldItalian",
    1657: u"OldSlavonic",  # x6
    32778: u"OldSpanish",
    1096: u"Oriya",
    1547: u"Ossetic",
    1145: u"Papiamento",  # x6
    1624: u"Papiamento_Legacy",
    1625: u"PidginEnglish",
    1654: u"Pinyin",
    1045: u"Polish",
    1046: u"Portuguese",            # not supported
    # 1046: u"PortugueseBrazilian",
    2070: u"PortugueseStandard",    # x5 supported
    1541: u"Provencal",
    1094: u"Punjabi",
    1131: u"Quechua",  # x6
    # 1131: u"QuechuaBolivia",  # x6
    2155: u"QuechuaEcuador",  # x6
    3179: u"QuechuaPeru",  # x6
    1626: u"Quechua_Legacy",
    1047: u"RhaetoRomanic",
    1048: u"Romanian",
    2072: u"RomanianMoldavia",
    1627: u"Romany",
    1628: u"Ruanda",
    1629: u"Rundi",
    1049: u"Russian",
    2073: u"RussianMoldavia",
    32793: u"RussianOldSpelling",
    34841: u"RussianOldOrtho",  # x6
    33817: u"RussianProperNames",
    1083: u"Saami",
    1542: u"Samoan",
    1103: u"Sanskrit",
    1630: u"Selkup",
    3098: u"SerbianCyrillic",
    2074: u"SerbianLatin",
    1631: u"Shona",
    1113: u"Sindhi",
    1051: u"Slovak",
    1060: u"Slovenian",
    1143: u"Somali",  # x6
    1633: u"Somali_Legacy",
    1070: u"Sorbian",                   # not supported
    1634: u"Sotho",
    # 1034: u"Spanish",                  # not supported
    1034: u"SpanishTraditionalSort",    # x5 tested
    11274: u"SpanishArgentina",
    16394: u"SpanishBolivia",
    13322: u"SpanishChile",
    9226: u"SpanishColombia",
    5130: u"SpanishCostaRica",
    7178: u"SpanishDominicanRepublic",
    12298: u"SpanishEcuador",
    17418: u"SpanishElSalvador",
    4106: u"SpanishGuatemala",
    18442: u"SpanishHonduras",
    2058: u"SpanishMexican",
    3082: u"SpanishModernSort",
    19466: u"SpanishNicaragua",
    6154: u"SpanishPanama",
    15370: u"SpanishParaguay",
    10250: u"SpanishPeru",
    33802: u"SpanishProperNames",
    20490: u"SpanishPuertoRico",
    14346: u"SpanishUruguay",
    8202: u"SpanishVenezuela",
    1635: u"Sunda",
    1072: u"Sutu",
    1089: u"Swahili",
    1636: u"Swazi",
    1053: u"Swedish",
    2077: u"SwedishFinland",
    1637: u"Tabassaran",
    1553: u"Tagalog",
    1639: u"Tahitian",
    1064: u"Tajik",  # x6
    1638: u"Tajik_Legacy",
    1097: u"Tamil",
    1092: u"Tatar",
    1098: u"Telugu",
    1054: u"Thai",
    1105: u"Tibet",
    1641: u"Tongan",
    1073: u"Tsonga",
    1074: u"Tswana",
    # 1074, u"Chuana",
    1642: u"Tun",
    1055: u"Turkish",
    1090: u"Turkmen",  # x6
    1643: u"Turkmen_Legacy",
    1644: u"Tuvin",
    1645: u"Udmurt",
    # 1646: u"Uighur",           # not supported
    1646: u"UighurCyrillic",    # not supported
    1647: u"UighurLatin",
    1058: u"Ukrainian",
    1653: u"Universal",
    2080: u"UrduIndia",
    1056: u"UrduPakistan",
    1554: u"User",
    2115: u"UzbekCyrillic",
    1091: u"UzbekLatin",
    1075: u"Venda",
    1066: u"Vietnamese",
    1648: u"Visayan",
    1106: u"Welsh",  # x6
    1543: u"Welsh_Legacy",
    1160: u"Wolof",  # x6
    1649: u"Wolof_Legacy",
    1076: u"Xhosa",
    1157: u"Yakut",  # x6
    1650: u"Yakut_Legacy",
    1085: u"Yiddish",
    1651: u"Zapotec",
    1077: u"Zulu",
}


def int2unichr(value):
    # import sys
    # if sys.version_info > (3, 0):
    #     return chr(value)
    # else:
    #     return unichr(value)
    return text(value)


def print_codecs():
    # sort by values
    for lcid in sorted(lang_map, key=lang_map.get):
        print("%6d:\t%s" % (lcid, lang_map[lcid]))


def bit_length(num):
    res = 1
    num >>= 1
    while num != 0:
        res += 1
        num >>= 1
    return res


def display_time(sec):
    result = ""
    hour = sec // 3600
    if hour:
        result += "{} hours".format(hour)
        if hour == 1:
            result = result.rstrip('s')
    sec -= 3600 * hour
    min = sec // 60
    if min:
        if result != "":
            result += " "
        result += "{} min".format(int(min))
    sec -= 60 * min
    if result != "":
        result += " "
    result += "{0:0.2f} sec".format(sec)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (print_function)
from . import tools

__author__ = 'sv99'


class SymInfo:
    def __init__(self, sym_idx, size, code):
        self.sym_idx = sym_idx
        self.size = size
        self.code = code


class HuffmanNode:
    def __init__(self, left, right, parent, weight):
        self.left = left
        self.right = right
        self.parent = parent
        self.weight = weight


class LenTable:
    def __init__(self, bstr):
        self.bstr = bstr
        self._count = self.bstr.read_bits(32)
        self._bits_per_len = self.bstr.read_bits(8)
        self._idx_bit_size = tools.bit_length(self._count)

        self.symidx2nodeidx = [-1 for _ in range(self._count)]
        self.nodes = [HuffmanNode(0, 0, -1, -1) for _ in range(self._count - 1)]
        root_idx = len(self.nodes) - 1
        self.next_node_position = 0
        for i in range(self._count):
            symidx = self.bstr.read_bits(self._idx_bit_size)
            length = self.bstr.read_bits(self._bits_per_len)
            self.place_sym_idx(symidx, root_idx, length)

    def place_sym_idx(self, sym_idx, node_idx, size):
        assert size > 0
        if size == 1:  # time to place
            if self.nodes[node_idx].left == 0:
                self.nodes[node_idx].left = -1 - sym_idx
                self.symidx2nodeidx[sym_idx] = node_idx
                return True

            if self.nodes[node_idx].right == 0:
                self.nodes[node_idx].right = -1 - sym_idx
                self.symidx2nodeidx[sym_idx] = node_idx
                return True

            return False

        if self.nodes[node_idx].left == 0:
            self.nodes[self.next_node_position] = HuffmanNode(0, 0, node_idx, -1)
            self.next_node_position += 1
            self.nodes[node_idx].left = self.next_node_position

        if self.nodes[node_idx].left > 0:
            if self.place_sym_idx(sym_idx, self.nodes[node_idx].left - 1, size - 1):
                return True

        if self.nodes[node_idx].right == 0:
            self.nodes[self.next_node_position] = HuffmanNode(0, 0, node_idx, -1)
            self.next_node_position += 1
            self.nodes[node_idx].right = self.next_node_position

        if self.nodes[node_idx].right > 0:
            if self.place_sym_idx(sym_idx, self.nodes[node_idx].right - 1, size - 1):
                return True

        return False

    def decode(self):
        node = self.nodes[-1]
        length = 0
        while True:
            length += 1
            bit = self.bstr.read_bit()
            if bit:  # right
                if node.right < 0:  # leaf
                    sym_idx = -1 - node.right
                    return sym_idx
                node = self.nodes[node.right - 1]
            else:  # left
                if node.left < 0:  # leaf
                    sym_idx = -1 - node.left
                    return sym_idx

                node = self.nodes[node.left - 1]

    def dump(self, name):
        print("LenTable:              %s" % name)
        print("    Count:             %s" % self._count)
        print("    bitsPerLen:        %s" % self._bits_per_len)
        print("    IdxBitSize:        %s" % self._idx_bit_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import struct

from lingvoreader.tools import int2unichr
from lingvoreader import LsdError

__author__ = 'sv99'


def reverse32(int_value):
    res, = struct.unpack('>L', struct.pack('<L', int_value))
    return res


def reverse16(word_value):
    res, = struct.unpack('>H', struct.pack('<H', word_value))
    return res


def rec2word(rec, big_endian=True):
    if big_endian:
        return (ord(rec[0]) << 8) + ord(rec[1])
    else:
        return (ord(rec[1]) << 8) + ord(rec[0])


def rec2int(rec, big_endian=True):
    if big_endian:
        return (ord(rec[0]) << 24) + (ord(rec[1]) << 16) + (ord(rec[2]) << 8) + ord(rec[3])
    else:
        return (ord(rec[3]) << 24) + (ord(rec[2]) << 16) + (ord(rec[1]) << 8) + ord(rec[0])


class BitStream:
    def __init__(self, record):
        """ record - bytearray! """
        self.record = record
        self.pos = 0
        self.in_byte_pos = 0  # start from highest to smallest
        return

    @property
    def length(self):
        return len(self.record)

    def seek(self, pos):
        self.pos = pos
        self.in_byte_pos = 0
        return pos < self.length

    def read(self, length):
        current = self.pos
        self.pos += length
        return self.record[current:self.pos]

    def to_nearest_byte(self):
        """
        move to start next byte, if needed
        """
        if self.in_byte_pos != 0:
            self.in_byte_pos = 0
            self.pos += 1

    def read_byte(self):
        res, = struct.unpack_from('B', self.record, self.pos)
        self.pos += 1
        self.in_byte_pos = 0
        return res

    # def read_word_rev(self):
    #    res, = struct.unpack_from('<H', self.record, self.pos)
    #    self.pos += 2
    #    self.in_byte_pos = 0
    #    return res
    #
    # def read_int_rev(self):
    #    res, = struct.unpack_from('<L', self.record, self.pos)
    #    self.pos += 4
    #    self.in_byte_pos = 0
    #    return res

    def read_word(self):
        res, = struct.unpack_from('>H', self.record, self.pos)
        self.pos += 2
        self.in_byte_pos = 0
        return res

    def read_int(self):
        res, = struct.unpack_from('>L', self.record, self.pos)
        self.pos += 4
        self.in_byte_pos = 0
        return res

    def read_symbols(self):
        size = self.read_bits(32)
        bits_per_symbol = self.read_bits(8)
        res = []
        for i in range(size):
            res.append(self.read_bits(bits_per_symbol))
        return res

    def read_bit(self):
        byte, = struct.unpack_from('B', self.record, self.pos)
        byte >>= (7 - self.in_byte_pos)
        if self.in_byte_pos == 7:
            self.pos += 1
            self.in_byte_pos = 0
        else:
            self.in_byte_pos += 1
        return byte & 1

    def read_bits(self, count):
        return self.read_bits_o(count)

    # stupid direct implementation
    def read_bits_s(self, count):
        if count > 32:
            raise LsdError("Many bits for read: %d" % count)
        res = 0
        for i in range(count):
            res <<= 1
            res += self.read_bit()
        return res

    def read_bits_o(self, count):
        if count > 32:
            raise LsdError("Many bits for read: %d" % count)
        masks = (1, 3, 7, 0xF, 0x1F, 0x3F, 0x7F, 0xFF)
        count_bytes = (count + self.in_byte_pos) // 8
        if count + self.in_byte_pos - 8 * count_bytes > 0:
            count_bytes += 1
        # if in single raw byt
        if count_bytes == 1:
            if (self.in_byte_pos + count) < 8:
                byte = self.record[self.pos]
                byte >>= 7 - self.in_byte_pos - count + 1
                byte &= masks[count - 1]
                self.in_byte_pos += count
                return byte
        # many raw bytes
        #   inBitPos
        #      |   count = 13    |
        # 01234567 | 01234567 | 0123456
        #
        # inBitPos = 5 count_first = 3 count_las = 2
        #
        p = self.pos
        count_last = (count + self.in_byte_pos) % 8
        count_first = 8 - self.in_byte_pos
        byte_first = self.record[p]
        p += 1
        byte_first &= masks[count_first - 1]
        res = byte_first
        # full bytes
        full_bytes = (count - count_first) // 8
        if full_bytes > 0:
            for i in range(full_bytes):
                res <<= 8
                res += self.record[p]
                p += 1
        # last byte
        if count_last > 0:
            byte = self.record[p]
            byte >>= 8 - count_last
            res <<= count_last
            res += byte
        self.in_byte_pos = count_last
        self.pos = p
        return res

    def read_some(self, length):
        if length == 1:
            return self.read_byte()
        elif length == 2:
            return self.read_word()
        elif length == 4:
            return self.read_int()
        else:
            raise LsdError('Allow read byte, word and int length: %i' % length)

    def read_unicode(self, size, big_endian=True):
        res = ""
        for i in range(size):
            ch = self.read_some(2)
            if not big_endian:
                ch = reverse16(ch)
            res += int2unichr(ch)
            # res += unichr(self.read_some(2), big_endian))
        return res





#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import (division, absolute_import, print_function, with_statement)

import argparse
import codecs
import os
import sys
from timeit import default_timer as timer

from lingvoreader import __version__
from lingvoreader import LsdError
from lingvoreader import tools
from lingvoreader.lsdfile import LsdFile

__author__ = 'sv99'

#
# lsd decoder - based on source from tr7 user from ru-board
#   http://forum.ru-board.com/profile.cgi?action=show&member=tr7
# ru-board forum - "Lingvo dictionary"
#   http://forum.ru-board.com/topic.cgi?forum=93&topic=3774&glp#lt
# current version on the github https://github.com/nongeneric/lsd2dsl
#
# Worked with Lingvo x5 dictionary, other version not supported
# v 0.1 - 16.09.2013
#
# v 0.2.0 - 22.11.2015
#   add support lingvo 10, 12, x3 and x6 dictionary format
#
# v 0.2.9 - 28.12.2018
#   add python3 support
#
if sys.platform.startswith("win"):
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout)


def unpack(dicts, dest_dir, verbose):
    # dict_ext = os.path.splitext(dict_file)[1].upper()
    # if dict_ext != '.LSD':
    #     raise LsdError("Need Lingvo lsd dictionary.")

    count = len(dicts)
    if count == 1:
        print("Unpacking dict: %s" % dicts[0])
    for i in range(count):
        dict_file = dicts[i]
        start = timer()
        try:
            if count > 1:
                print("Unpacking dict (%d from %d): %s" % (i + 1, count, dict_file))
            m = LsdFile(dict_file, verbose)
            m.parse()
            m.dump()
            m.write(dest_dir)
        except ValueError as e:
            print("Error: %s" % e)
            return 1
        end = timer()
        print("Unpack OK (%s)" % tools.display_time(end - start))

    return 0


def header(dicts):
    # dict_ext = os.path.splitext(dict_file)[1].upper()
    # if dict_ext != '.LSD':
    #     raise LsdError("Need Lingvo lsd dictionary.")

    count = len(dicts)
    if count == 1:
        print("Unpacking dict: %s" % dicts[0])
    for i in range(count):
        dict_file = dicts[i]
        try:
            if count > 1:
                print("Unpacking dict (%d from %d): %s" % (i + 1, count, dict_file))
            m = LsdFile(dict_file, True)
            m.dump()
            # print("Header %s OK" % dict_file)
        except ValueError as e:
            print("Error: %s" % e)
            return 1

    return 0


def get_dicts():
    current = os.getcwd()
    res = []
    for f in os.listdir(current):
        if f.endswith(".lsd"):
            res.append(f)
    return res


class CodecsAction(argparse.Action):
    def __init__(self,
                 option_strings,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help="print supported languages and their codes"):
        super(CodecsAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.exit(message=tools.print_codecs())


def get_arg_parser():
    p = argparse.ArgumentParser(description='Decode Lingvo lsd dictionary to dsl')
    g = p.add_mutually_exclusive_group(required=True)
    g.add_argument("-i", "--input", help='Dictionary to decode')
    g.add_argument("-a", "--all", action="store_true", help='All dictionary in current directory')
    p.add_argument("--header", action="store_true", default=False, help='Print dictionary header and exit')
    p.add_argument("-o", "--outdir", default="", help="Output directory")
    p.add_argument("-c", "--codecs", action=CodecsAction)
    p.add_argument("-v", "--verbose", action="store_true", default=False)
    p.add_argument('--version', action='version', version='%(prog)s ' + __version__)
    return p


def main():
    args = get_arg_parser().parse_args()
    dicts = []
    if args.all:
        # all lsd in directory
        print("Decode all lsd in current directory..")
        dicts = get_dicts()
        print(dicts)
    else:
        dicts.append(args.input)

    if args.header:
        header(dicts)
    else:
        if args.outdir != "":
            # check path
            if not os.path.exists(args.outdir):
                os.mkdir(args.outdir)

        start = timer()
        unpack(dicts, args.outdir, args.verbose)
        end = timer()
        if len(dicts) > 1:
            # print("Files count: %i" % c)
            print("Elapsed: %s" % tools.display_time(end - start))

    return 0


if __name__ == '__main__':
    sys.exit(main())



# coding: utf-8
from __future__ import unicode_literals, print_function, division, absolute_import

import codecs
import os

from lingvoreader import LsdError
from lingvoreader import tools, decoder
from lingvoreader.articleheading import ArticleHeading, ArticleHeadingList
from lingvoreader.bitstream import reverse32, reverse16, BitStream

__author__ = 'sv99'


class OverlayReader:
    def __init__(self, bstr, offset):
        self.bstr = bstr

        if self.bstr.seek(offset):
            self._entriesCount = self.bstr.read_bits(4)
        else:
            self._entriesCount = 0

    def dump(self):
        print("Overlay:")
        print("    EntriesCount:      %s" % self._entriesCount)


class Header:
    def __init__(self, bstr):
        self.bstr = bstr
        self.magic = self.bstr.read(8).decode().replace('\x00', '')
        self.version = reverse32(self.bstr.read_int())
        self.unk = reverse32(self.bstr.read_int())
        self.checksum = reverse32(self.bstr.read_int())
        self.entries_count = reverse32(self.bstr.read_int())
        self.annotation_offset = reverse32(self.bstr.read_int())
        self.dictionary_encoder_offset = reverse32(self.bstr.read_int())
        self.articles_offset = reverse32(self.bstr.read_int())
        self.pages_offset = reverse32(self.bstr.read_int())
        self.unk1 = reverse32(self.bstr.read_int())
        self.unk2 = reverse16(self.bstr.read_word())
        self.unk3 = reverse16(self.bstr.read_word())
        self.source_language = reverse16(self.bstr.read_word())
        self.target_language = reverse16(self.bstr.read_word())
        return

    @property
    def hi_version(self):
        return self.version >> 16

    def dump(self):
        print("Header:")
        print("    Magic:             %s" % self.magic)
        print("    Checksume:         %s" % hex(self.checksum))
        print("    Version:           %s (%s)" % (hex(self.hi_version), hex(self.version)))
        print("    Entries:           %d" % self.entries_count)
        print("    AnnotationOffset:  %s" % hex(self.annotation_offset))
        print("    DictionaryEncoderOffset: %s" % hex(self.dictionary_encoder_offset))
        print("    ArticlesOffset:    %s" % hex(self.articles_offset))
        print("    Pages start:       %s" % hex(self.pages_offset))
        print("    Source language:   %d %s" % (self.source_language, tools.lang_map[self.source_language]))
        print("    Target language:   %d %s" % (self.target_language, tools.lang_map[self.target_language]))


class CachePage:
    def __init__(self, bstr):
        self.bstr = bstr
        self.is_leaf = bstr.read_bit()
        self.number = bstr.read_bits(16)
        self.prev = bstr.read_bits(16)
        self.parent = bstr.read_bits(16)
        self.next = bstr.read_bits(16)
        self.headings_count = bstr.read_bits(16)
        self.bstr.to_nearest_byte()
        return


xor_pad = (
    0x9C, 0xDF, 0x9B, 0xF3, 0xBE, 0x3A, 0x83, 0xD8,
    0xC9, 0xF5, 0x50, 0x98, 0x35, 0x4E, 0x7F, 0xBB,
    0x89, 0xC7, 0xE9, 0x6B, 0xC4, 0xC8, 0x4F, 0x85,
    0x1A, 0x10, 0x43, 0x66, 0x65, 0x57, 0x55, 0x54,
    0xB4, 0xFF, 0xD7, 0x17, 0x06, 0x31, 0xAC, 0x4B,
    0x42, 0x53, 0x5A, 0x46, 0xC5, 0xF8, 0xCA, 0x5E,
    0x18, 0x38, 0x5D, 0x91, 0xAA, 0xA5, 0x58, 0x23,
    0x67, 0xBF, 0x30, 0x3C, 0x8C, 0xCF, 0xD5, 0xA8,
    0x20, 0xEE, 0x0B, 0x8E, 0xA6, 0x5B, 0x49, 0x3F,
    0xC0, 0xF4, 0x13, 0x80, 0xCB, 0x7B, 0xA7, 0x1D,
    0x81, 0x8B, 0x01, 0xDD, 0xE3, 0x4C, 0x9A, 0xCE,
    0x40, 0x72, 0xDE, 0x0F, 0x26, 0xBD, 0x3B, 0xA3,
    0x05, 0x37, 0xE1, 0x5F, 0x9D, 0x1E, 0xCD, 0x69,
    0x6E, 0xAB, 0x6D, 0x6C, 0xC3, 0x71, 0x1F, 0xA9,
    0x84, 0x63, 0x45, 0x76, 0x25, 0x70, 0xD6, 0x8F,
    0xFD, 0x04, 0x2E, 0x2A, 0x22, 0xF0, 0xB8, 0xF2,
    0xB6, 0xD0, 0xDA, 0x62, 0x75, 0xB7, 0x77, 0x34,
    0xA2, 0x41, 0xB9, 0xB1, 0x74, 0xE4, 0x95, 0x1B,
    0x3E, 0xE7, 0x00, 0xBC, 0x93, 0x7A, 0xE8, 0x86,
    0x59, 0xA0, 0x92, 0x11, 0xF7, 0xFE, 0x03, 0x2F,
    0x28, 0xFA, 0x27, 0x02, 0xE5, 0x39, 0x21, 0x96,
    0x33, 0xD1, 0xB2, 0x7C, 0xB3, 0x73, 0xC6, 0xE6,
    0xA1, 0x52, 0xFB, 0xD4, 0x9E, 0xB0, 0xE2, 0x16,
    0x97, 0x08, 0xF6, 0x4A, 0x78, 0x29, 0x14, 0x12,
    0x4D, 0xC1, 0x99, 0xBA, 0x0D, 0x3D, 0xEF, 0x19,
    0xAF, 0xF9, 0x6F, 0x0A, 0x6A, 0x47, 0x36, 0x82,
    0x07, 0x9F, 0x7D, 0xA4, 0xEA, 0x44, 0x09, 0x5C,
    0x8D, 0xCC, 0x87, 0x88, 0x2D, 0x8A, 0xEB, 0x2C,
    0xB5, 0xE0, 0x32, 0xAD, 0xD3, 0x61, 0xAE, 0x15,
    0x60, 0xF1, 0x48, 0x0E, 0x7E, 0x94, 0x51, 0x0C,
    0xEC, 0xDB, 0xD2, 0x64, 0xDC, 0xFC, 0xC2, 0x56,
    0x24, 0xED, 0x2B, 0xD9, 0x1C, 0x68, 0x90, 0x79
)


class LsdFile:
    def __init__(self, dict_file, verbose=False):
        self.filename = dict_file
        self._readed = False
        self._parsed = False
        self.verbose = verbose
        with open(dict_file, 'rb') as fp:
            self.bstr = BitStream(bytearray(fp.read()))

        self.overlay = None
        self.headings = ArticleHeadingList()
        self.dict = []
        self.header = Header(self.bstr)
        # check magic
        if self.header.magic != u'LingVo':
            raise LsdError('Allow only Lsd "LingVo" ident: %s' % repr(self.header.magic))

        # initialize decoder
        self.decoder = None
        hi_version = self.header.hi_version
        version = self.header.version
        if hi_version == 0x11:  # lingvo 11 dictionary: 0x11001
            self.decoder = decoder.UserDictionaryDecoder(self.bstr)
        elif hi_version == 0x12:  # lingvo 12 dictionary: 0x12001
            self.decoder = decoder.UserDictionaryDecoder(self.bstr)
        elif hi_version == 0x13:  # x3 dictionary: 0x131001 and 0x132001 if pages count > 1000
            self.decoder = decoder.SystemDictionaryDecoder13(self.bstr)
        elif hi_version == 0x14:  # x5 dictionary
            if version == 0x142001:  # user dictionaries
                self.decoder = decoder.UserDictionaryDecoder(self.bstr)
            elif version == 0x141004:  # system dictionaries
                self.decoder = decoder.SystemDictionaryDecoder14(self.bstr)
            elif version == 0x145001:  # abbreviation dictionaries
                self.decoder = decoder.AbbreviationDictionaryDecoder(self.bstr)
        elif hi_version == 0x15:  # x6 dictionary
            if version == 0x152001:  # user dictionaries
                self.decoder = decoder.UserDictionaryDecoder(self.bstr)
            elif version == 0x151005:  # system dictionaries
                # xor dictionary
                self.xor_block_x6(self.header.dictionary_encoder_offset, self.header.articles_offset)
                self.decoder = decoder.SystemDictionaryDecoder14(self.bstr)
            elif version == 0x155001:  # abbreviation dictionaries
                self.decoder = decoder.AbbreviationDictionaryDecoder(self.bstr)

        if self.decoder is None:
            self.dump()
            print("Not supported dictionary version: %s" % hex(self.header.version))
            exit(1)
            # raise LsdError("Not supported dict version %s" % hex(self.header.version))

        name_len = self.bstr.read_some(1)
        self.name = self.bstr.read_unicode(name_len, False)
        self.first_heading = self.bstr.read_unicode(self.bstr.read_byte(), False)
        self.last_heading = self.bstr.read_unicode(self.bstr.read_byte(), False)
        capitals_len = reverse32(self.bstr.read_int())
        self.capitals = self.bstr.read_unicode(capitals_len, False)
        # icon v12+
        if self.header.version > 0x120000:
            self.icon_size = reverse16(self.bstr.read_word())
            self.icon = self.bstr.read(self.icon_size)
        else:
            self.icon_size = 0
            self.icon = None

        if self.header.version > 0x140000:
            self.header_checksum = reverse32(self.bstr.read_int())
        else:
            self.header_checksum = 0

        if self.header.version > 0x120000:
            self.pages_end = reverse32(self.bstr.read_int())
            self.overlay_data = reverse32(self.bstr.read_int())
        else:
            self.pages_end = self.bstr.length
            self.overlay_data = self.bstr.length  # no overlay

        if self.header.version > 0x140000:
            self.dummy1 = reverse32(self.bstr.read_int())
            self.dummy2 = reverse32(self.bstr.read_int())
        else:
            self.dummy1 = 0
            self.dummy2 = 0

        # set bstr pos for decoding
        self.bstr.seek(self.header.dictionary_encoder_offset)

    # x6 system dictionary table based xor decoding
    # each block xored with start key=0x7f
    # 1. dictionary_encoder_offset -> article_offset
    #    must by decoded befor decoder.read()
    # 2. annotation_offset -> dictionary_encoder_offset
    #    annotation decoded in the read_annotation
    # 3. each article encoded individully
    #    articles_offset + heading.reference -> articles_offset + heading.next-reference
    #    article decoded in the
    def xor_block_x6(self, start, end, key=0x7f):
        for i in range(start, end):
            byte = self.bstr.record[i]
            self.bstr.record[i] = byte ^ key
            key = xor_pad[byte]
        return key

    @property
    def pages_count(self):
        return (self.pages_end - self.header.pages_offset) // 512

    def get_page_offset(self, page_number):
        return self.header.pages_offset + 512 * page_number

    def read_headings(self):
        for i in range(self.pages_count):
            self.read_heading_from_page(i)
        # set last next_reference
        self.headings[-1].next_reference = self.header.pages_offset - self.header.articles_offset

    def merge_headings(self):
        res = []
        # fill next_reference in the headings
        prev = self.headings[0]
        res.append(prev)
        for i in range(1, len(self.headings)):
            h = self.headings[i]
            if prev.reference == h.reference:
                # multititle article
                prev.merge(h)
            else:
                res[-1].next_reference = h.reference
                res.append(h)
            prev = h
        # headings[i].next_reference = headings[i+1].reference
        # set next_reference for last item to the pages_offset
        res[-1].next_reference = self.header.pages_offset - self.header.articles_offset
        return res

    def read_heading_from_page(self, page_number):
        self.bstr.seek(self.get_page_offset(page_number))
        page = CachePage(self.bstr)
        if page.is_leaf:
            prefix = ""
            for idx in range(page.headings_count):
                heading = ArticleHeading()
                prefix = heading.read(self.decoder, self.bstr, prefix)
                self.headings.append(heading)

    def read_article(self, heading):
        self.bstr.seek(self.header.articles_offset + heading.reference)
        if self.header.version == 0x151005:
            # xor article
            self.xor_block_x6(self.header.articles_offset + heading.reference,
                              self.header.articles_offset + heading.next_reference)
        size = self.bstr.read_bits(16)
        if size == 0xFFFF:
            size = self.bstr.read_bits(32)

        res = self.decoder.decode_article(size)
        # assert(res)
        return res

    def read_annotation(self):
        if self.header.version == 0x151005:
            # xor annotation
            self.xor_block_x6(self.header.annotation_offset,
                              self.header.dictionary_encoder_offset)
        res = ""
        if self.bstr.seek(self.header.annotation_offset):
            size = self.bstr.read_bits(16)
            res = self.decoder.decode_article(size)
        return res

    @property
    def readed(self):
        return self._readed

    def read(self):
        if self.verbose:
            print("reading dictionary..")
        self.decoder.read()
        self._readed = True

    @property
    def parsed(self):
        return self._parsed

    def parse(self):
        if not self.readed:
            self.read()
        if self.verbose:
            print("decoding overlay..")
        self.overlay = OverlayReader(self.bstr, self.overlay_data)

        if self.verbose:
            print("decoding headings: %d" % self.header.entries_count)
        self.read_headings()
        if self.headings.appended != self.header.entries_count:
            raise LsdError("Decoded not all entries %d != %d" % (self.headings.appended, self.header.entries_count))
        # merge multititle headings
        # self.headings = self.merge_headings()

        if self.verbose:
            print("decoding articles: %d" % len(self.headings))
        for h in self.headings:
            # h.dump()
            self.dict.append((h, self.read_article(h)))
        self._parsed = True
        if self.verbose:
            print("OK")

    def write(self, path=""):
        """ save decoded dictionary """
        if not self.parsed:
            self.parse()
        self.write_icon(path)
        self.write_annotation(path)
        self.write_overlay(path)
        self.write_dsl(path)
        if self.verbose:
            self.write_prefix(path)

    def make_filename(self, path, ext):
        base, orig_ext = os.path.splitext(self.filename)
        if path != "":
            base = os.path.join(path, os.path.basename(base))
        return base + '.' + ext

    def write_icon(self, path=""):
        if self.icon_size == 0:
            return
        ico_file = self.make_filename(path, "bmp")
        with open(ico_file, 'wb') as ico:
            ico.write(self.icon)
        if self.verbose:
            print('Write icon:       %s' % ico_file)

    def write_annotation(self, path=""):
        annotation = self.read_annotation()
        if annotation == "":
            return
        ann_file = self.make_filename(path, "ann")
        with codecs.open(ann_file, 'w', encoding='utf-16') as ann:
            ann.write(annotation)
        if self.verbose:
            print('Write annotation: %s' % ann_file)

    def write_prefix(self, path=""):
        if self.decoder.prefix == "":
            return
        pref_file = self.make_filename(path, "pref")
        with codecs.open(pref_file, 'w', encoding='utf-8') as pref:
            pref.write(self.decoder.prefix)
        if self.verbose:
            print('Write prefix:     %s' % pref_file)

    def write_overlay(self, path=""):
        pass

    @staticmethod
    def normalize_article(article):
        res = article.replace(u'\n', u'\n\t')
        return res

    def write_dsl(self, path=""):
        if len(self.dict) == 0:
            print("Nothing writing to dsl!")
            return
        dsl_file = self.make_filename(path, "dsl")
        with codecs.open(dsl_file, 'w', encoding='utf-16') as dsl:
            dsl.write(u"#NAME\t\"" + self.name + u"\"\n")
            dsl.write(u"#INDEX_LANGUAGE\t\"" + tools.lang_map[self.header.source_language] + u"\"\n")
            dsl.write(u"#CONTENTS_LANGUAGE\t\"" + tools.lang_map[self.header.target_language] + u"\"\n")
            if self.icon_size > 0:
                base, orig_ext = os.path.splitext(os.path.basename(self.filename))
                dsl.write(u"#ICON_FILE\t\"" + base + '.' + "bmp" + u"\"\n")
            dsl.write(u"\n")
            for h, r in self.dict:
                if h.simple:
                    dsl.write(h.get_first_ext_text())
                    dsl.write(u"\n\t")
                else:
                    for item in h.headings:
                        dsl.write(item.ext_text)
                        dsl.write(u"\n")
                    dsl.write(u"\t")
                dsl.write(self.normalize_article(r))
                dsl.write(u"\n")
        if self.verbose:
            print('Write dsl:        %s' % dsl_file)

    def dump(self):
        self.header.dump()
        # dump header for not supported versions
        if self.decoder is not None:
            print("Name:                  %s" % self.name)
            print("First heading:         %s" % self.first_heading)
            print("Last heading:          %s" % self.last_heading)
            print("Capitals:              %s" % self.capitals)
            print("Pages end:             %s" % hex(self.pages_end))
            print("Overlay data:          %s" % hex(self.overlay_data))
            print("Pages count:           %d" % ((self.pages_end - self.header.pages_offset) // 512))
            if self.header.version > 0x140000:
                print("dummy1:                %s" % hex(self.dummy1))
                print("dummy2:                %s" % hex(self.dummy2))
            print("Icon enable:           %s" % (self.icon_size > 0))
            if self.readed:
                self.decoder.dump()
                self.overlay.dump()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import os
import re

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(os.path.join(here, *parts), 'r').read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="lingvoreader",
    version=find_version("lingvoreader", "__init__.py"),
    author='sv99',
    author_email='sv99@inbox.ru',
    url='https://github.com/sv99/lsdreader',
    description='Linvo 11, 12, X3, X5 and X6 lsd reader Utilities\n compatible with python2.7 and python3',
    long_description=read('README.rst'),
    packages=find_packages(),
    platforms='any',
    install_requires=[
        'future; python_version == "2.7"',
    ],
    zip_safe=True,
    entry_points={
        'console_scripts': [
            'lsdreader = lingvoreader.lsdreader:main',
        ]
    }
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from unittest import TestCase
from lingvoreader import tools

__author__ = 'sv99'


class TestTools(TestCase):
    def test_bit_length_0(self):
        self.assertEqual(tools.bit_length(0), 1)

    def test_bit_length_1(self):
        self.assertEqual(tools.bit_length(1), 1)

    def test_bit_length_2(self):
        self.assertEqual(tools.bit_length(2), 2)

    def test_bit_length_4(self):
        self.assertEqual(tools.bit_length(4), 3)

    def test_display_time(self):
        self.assertEquals(tools.display_time(1),
                          "1.00 sec")
        self.assertEquals(tools.display_time(61.2),
                          "1 min 1.20 sec")
        self.assertEquals(tools.display_time(60),
                          "1 min 0.00 sec")
        self.assertEquals(tools.display_time(3600),
                          "1 hour 0.00 sec")
        self.assertEquals(tools.display_time(3661),
                          "1 hour 1 min 1.00 sec")



# coding: utf-8
from __future__ import unicode_literals, print_function, division, absolute_import
from lingvoreader.tools import int2unichr

__author__ = 'svolkov'


# info from ArticleHeading.cpp
# Not implemented yet!!
#
#      Unsorted parts
#
# Each characted in a LSD heading has a flag indicating if it's 'sorted' or 'unsorted'.
# The unsorted characters aren't used in sortings and searching. Basically they become
# transparent for the indexing mechanism and only visible while presented to the user
# together with an article. Let's annotate each character with an 's' or 'u' to indicate
# this distinction.
#
# In the DSL format curly brackets are used to denote the sortedness of a heading's part.
#
# The heading 'Some{thing}' would be encoded as
#
#   Something
#   ssssuuuuu
#
# And it's name would be just 'Some', while its extended name would be 'Something'
# There can be any number of unsorted characters in a heading. At any position.
#
# To correctly decompile an LSD heading (where each character is either sorted or unsorted)
# to a DSL heading, it is required to group adjacent characters in groups by their
# sortedness, and then enclose each such group in parentheses. This process is rather
# straightforward. The special case to look for is the slash (/), which is encoded
# as an unsorted character, but is used to escape the subsequent characted (either sorted
# or unsorted) and thus requires a special handling.
#
#      Optional parts (variants)
#
# The DSL format has a mechanism for generating several headings with optional parts
# out of a single heading. By enclosing a part of a heading in parentheses, it is possible
# to generate all possible combinations of the heading.
#
# The heading 'aa(bb)cc' would be expanded into two different headings
#
#   aa{(bb)}cc  and  aa{(}bb{)}cc
#   ss uuuu ss       ss u ss u ss
#
# For more such parts, more headings would be generated. It is 2 headings for 1 part,
# 4 headings for 2 parts, 9 for 3 and so on.
#
# There exists a pattern which helps to combine two headings produced with the variant
# encoding. The pattern looks like this 'u(' denotes an unsorted parenthesis):
#
#   (A)  ??? u(   u*n u) ???   ->   ??? s( s|u*n s) ???
#   (B)  ??? u( s|u*n u) ???
#
# An optional part can contain an unsorted part as well, so this pattern allow for a
# combination of sorted/unsorted parts in an optional part (denoted as 's|u*n').
#
# This pattern becomes insufficient in the presence of spaces, adjacent to the
# parentheses. Here are two examples of such headings (the upper case is used
# to denote spaces):
#
#   abc (123)
#
#   abc{ (123)}    and   abc {(}123{)}
#   sss Uuuuuu           sssS u sss u
#
#   bbb (123) z
#
#   bbb {(123) }z  and   bbb {(}123{)} z
#   sssS uuuuuU s        sssS u sss u Ss
#
# Another two patterns account for these special cases.
#
#   (C) ??? Uu(   u*n u)  ->  ??? Ss( s|u*n s)
#   (D) ??? Su( s|u*n u)
#
#   (E) ??? Su(   u*n u)U ???  ->  ??? Ss( s|u*n s)S ???
#   (F) ??? Su( s|u*n u)S ???
#
# The headings that can't be collapsed using one of these three rules are left as is.
#
#
class CharInfo:
    def __init__(self):
        self.sorted = False
        self.escaped = False
        self.chr = ""

    # add equal test


class Heading:
    def __init__(self):
        self.text = ""
        self.extensions = []

    @property
    def ext_text(self):
        if len(self.extensions) == 0:
            return self.text
        res = self.text
        offset = 0
        for idx, ext in self.extensions:
            add_braces = ext != u"\\"
            if add_braces:
                ext = u"{" + ext + u"}"

            res = res[:idx + offset] + ext + res[idx + offset:]
            if add_braces:
                offset += 2
        return res


class ArticleHeading:
    def __init__(self):
        # may by many heading for single article
        self.headings = []
        self.reference = 0
        # refernce for next arcticle, used in the x6 for encoding
        self.next_reference = 0

    def read(self, lsd_decoder, bstr, known_prefix):
        h = Heading()
        prefix_len = lsd_decoder.decode_prefix_len()
        postfix_len = lsd_decoder.decode_postfix_len()
        h.text = known_prefix[:prefix_len]
        h.text += lsd_decoder.decode_heading(postfix_len)
        self.reference = lsd_decoder.read_reference2()
        if bstr.read_bit():
            # additional not visible formatting item in header
            # join multisymbols item
            ext_length = bstr.read_bits(8)
            if ext_length != 0:
                ext = ""
                first_idx = prev_idx = 0
                for i in range(ext_length):
                    idx = bstr.read_bits(8)
                    char = int2unichr(bstr.read_bits(16))
                    if ext == "":
                        ext += char
                        first_idx = prev_idx = idx
                    else:
                        if prev_idx + 1 == idx:
                            # join item with sequence idx
                            ext += char
                            prev_idx = idx
                        else:
                            # other item
                            h.extensions.append((first_idx, ext))
                            ext = char
                            first_idx = prev_idx = idx
                # add last item
                h.extensions.append((first_idx, ext))
        self.headings.append(h)
        return h.text

    def merge(self, heading):
        for h in heading.headings:
            # print("merge: %s" % h.ext_text)
            self.headings.append(h)

    def get_first(self):
        return self.headings[0]

    def get_first_ext_text(self):
        return self.headings[0].ext_text

    @property
    def simple(self):
        return len(self.headings) == 1

    def dump(self):
        print("%s: %d - %d" % (self.get_first().text, self.reference, self.next_reference))


class ArticleHeadingList(list):
    def __init__(self):
        super(ArticleHeadingList, self).__init__()
        self.appended = 0
        self.references = {}

    def append(self, item):
        self.appended += 1
        # append if item.reference not exists
        ref = item.reference
        if ref in self.references:
            self.references[ref].merge(item)
        else:
            self.references[ref] = item
            # if not first then update next_reference in the previous item
            if len(self) > 0:
                self[-1].next_reference = ref
            super(ArticleHeadingList, self).append(item)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from unittest import TestCase
from lingvoreader import lsdreader
import sys
from io import BytesIO

__author__ = 'sv99'


class ArgumentParserError(Exception):

    def __init__(self, message, stdout=None, stderr=None, error_code=None):
        Exception.__init__(self, message, stdout, stderr)
        self.message = message
        self.stdout = stdout
        self.stderr = stderr
        self.error_code = error_code


class StdIOBuffer(BytesIO):
    pass


def stderr_to_parser_error(parse_args, *args, **kwargs):
    # if this is being called recursively and stderr or stdout is already being
    # redirected, simply call the function and let the enclosing function
    # catch the exception
    if isinstance(sys.stderr, StdIOBuffer) or isinstance(sys.stdout, StdIOBuffer):
        return parse_args(*args, **kwargs)

    # if this is not being called recursively, redirect stderr and
    # use it as the ArgumentParserError message
    old_stdout = sys.stdout
    old_stderr = sys.stderr
    sys.stdout = StdIOBuffer()
    sys.stderr = StdIOBuffer()
    try:
        try:
            result = parse_args(*args, **kwargs)
            for key in list(vars(result)):
                if getattr(result, key) is sys.stdout:
                    setattr(result, key, old_stdout)
                if getattr(result, key) is sys.stderr:
                    setattr(result, key, old_stderr)
            return result
        except SystemExit:
            code = sys.exc_info()[1].code
            stdout = sys.stdout.getvalue()
            stderr = sys.stderr.getvalue()
            raise ArgumentParserError("SystemExit", stdout, stderr, code)
    finally:
        sys.stdout = old_stdout
        sys.stderr = old_stderr


class TestArgParser(TestCase):
    def setUp(self):
        self.parser = lsdreader.get_arg_parser()

    def test_empty(self):
        self.assertRaisesRegexp(ArgumentParserError,
                                'one of the arguments -i/--input -a/--all is required',
                                stderr_to_parser_error, self.parser.parse_args, [])

    def test_codecs(self):
        self.assertRaisesRegexp(ArgumentParserError,
                                "SystemExit",
                                stderr_to_parser_error, self.parser.parse_args, ['-c', ])

    def test_input(self):
        args = self.parser.parse_args('-i test'.split())
        self.assertEqual(args.input, 'test')
        self.assertFalse(args.verbose)

    def test_input_verbose(self):
        args = self.parser.parse_args('-i test -v'.split())
        self.assertEqual(args.input, 'test')
        self.assertTrue(args.verbose)
        self.assertEqual(args.outdir, '')

    def test_all(self):
        args = self.parser.parse_args('-a'.split())
        self.assertTrue(args.all)
        self.assertFalse(args.verbose)
        self.assertEqual(args.outdir, '')

    def test_all_verbose(self):
        args = self.parser.parse_args('-a -v'.split())
        self.assertTrue(args.all)
        self.assertTrue(args.verbose)
        self.assertEqual(args.outdir, '')

    def test_outdir(self):
        args = self.parser.parse_args('-a -o test'.split())
        self.assertTrue(args.all)
        self.assertFalse(args.verbose)
        self.assertEqual(args.outdir, 'test')
